#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include <netdb.h>
#include <sys/ioctl.h>
#include <mISDN/mISDNif.h>
#include "global.h"

void main_mport(int cardnr) {
#define BUFLEN 512
  static int dch_echo = 0; /*	  <<<===  	*/
  struct ctstamp {
    size_t	cmsg_len;
    int		cmsg_level;
    int		cmsg_type;
    struct	timeval	tv;
  };
  struct	sockaddr_mISDN  log_addr;
  char		time[16];
  struct	msghdr	mh;
  struct	iovec	iov[1];
  struct	ctstamp	cts;
  struct	tm	*mt;
  struct	mISDN_devinfo  di;
  struct	mISDNhead   *hh;
  struct	mISDNversion  ver;
  int		result, opt, channel, log_socket;
  u_int		cnt;
  u_char	buffer[BUFLEN];
#ifdef DEBUG
  int	i;
  char	*pn,	pns[43],	dir;
#endif

  if ((log_socket = socket(PF_ISDN, SOCK_RAW, 0)) < 0) {
    printf("could not open socket %s\n", strerror(errno));
  }

  result = ioctl(log_socket, IMGETVERSION, &ver);
  if (result < 0) {
    printf("ioctl error %s\n", strerror(errno));
    exit(10);
  }
  printf("mISDN kernel version %d.%02d.%d found\n", ver.major, ver.minor, ver.release);
  printf("mISDN user   version %d.%02d.%d found\n", MISDN_MAJOR_VERSION, MISDN_MINOR_VERSION, MISDN_RELEASE);

  if (ver.major != MISDN_MAJOR_VERSION) {
    printf("VERSION incompatible please update\n");
    exit(11);
  }

  result = ioctl(log_socket, IMGETCOUNT, &cnt);
  if (result < 0) {
    printf("ioctl error %s\n", strerror(errno));
    exit(12);
  }

  di.id = cardnr;
  result = ioctl(log_socket, IMGETDEVINFO, &di);
  if (result < 0) {
    printf("ioctl error %s\n", strerror(errno));
  } else {
#ifdef DEBUG
    printf("  id:    %d\n", di.id);
    printf("  Dprotocols:  %08x\n", di.Dprotocols);
    printf("  Bprotocols:  %08x\n", di.Bprotocols);
    printf("  protocol:  %d\n", di.protocol);
    printf("  channelmap:  ");
    for (i = MISDN_CHMAP_SIZE - 1; i >= 0; i--)
      printf("%02x", di.channelmap[i]);
    printf("\n");
    printf("  nrbchan:  %d\n", di.nrbchan);
#endif
    printf("Interface name: %s\n", di.name);
  }

  close(log_socket);



  if (di.protocol == ISDN_P_NONE) /* default TE */
    di.protocol = ISDN_P_TE_S0;

  if ((log_socket = socket(PF_ISDN, SOCK_DGRAM, di.protocol)) < 0) {
    printf("could not open log socket %s\n", strerror(errno));
    exit(1);
  }

  log_addr.family = AF_ISDN;
  log_addr.dev = cardnr;

  /* try to bind on D/E channel first, fallback to D channel on error */
  result = -1;
  channel = dch_echo;

  while ((result < 0) && (channel >= 0)) {
    log_addr.channel = (unsigned char)channel;
    result = bind(log_socket, (struct sockaddr *) &log_addr, sizeof(log_addr));
#ifdef DEBUG
    printf("log bind ch(%i) return %d\n", log_addr.channel, result);
#endif
      if (result < 0) {
      printf("log bind error %s\n", strerror(errno));
      close(log_socket);
      if (channel == 0) {
	exit(1);
      }
      channel--;
      if ((log_socket = socket(PF_ISDN, SOCK_DGRAM, di.protocol)) < 0) {
	printf("could not open log socket %s\n", strerror(errno));
	exit(1);
      }
    }
  }
  if (dch_echo)
    dch_echo = (log_addr.channel == 1);
  printf("Echo channel logging: %s\n", dch_echo ? "yes" : "no");

  opt = 1;
  result = setsockopt(log_socket, SOL_MISDN, MISDN_TIME_STAMP, &opt, sizeof(opt));
  if (result < 0) {
    printf("log  setsockopt error %s\n", strerror(errno));
  }


  hh = (struct mISDNhead *)buffer;

  while (1) {
    mh.msg_name = NULL;
    mh.msg_namelen = 0;
    mh.msg_iov = iov;
    mh.msg_iovlen = 1;
    mh.msg_control = &cts;
    mh.msg_controllen = sizeof(cts);
    mh.msg_flags = 0;
    iov[0].iov_base = buffer;
    iov[0].iov_len = BUFLEN;
    result = recvmsg(log_socket, &mh, 0);
    if (result < 0) {
      printf("read error %s\n", strerror(errno));
      break;
    } else {
      if (mh.msg_flags) {
	printf("received message with msg_flags(%x)\n", mh.msg_flags);
      }
      if (cts.cmsg_type == MISDN_TIME_STAMP) {
	mt = localtime((time_t *)&cts.tv.tv_sec);
	sprintf(time, "%02d:%02d:%02d.%06ld", mt->tm_hour, mt->tm_min, mt->tm_sec, cts.tv.tv_usec);
	time[12]=0;
      }
#ifdef DEBUG
      dir=-1;
      switch (hh->prim) {
	case PH_DATA_E_IND:		{pn="ECHO IND";			dir=1;}	break;
	case PH_DATA_IND:		{pn="DATA IND";			dir=1;}	break;
	case PH_DATA_REQ:		{pn="DATA REQ";			dir=0;}	break;
	case PH_DATA_CNF:		{pn="DATA CNF";			dir=2;}	break;
	case PH_ACTIVATE_IND:		{pn="ACTIVATE IND";		dir=1;}	break;
	case PH_ACTIVATE_REQ:		{pn="ACTIVATE REQ";		dir=0;}	break;
	case PH_ACTIVATE_CNF:		{pn="ACTIVATE CNF";		dir=2;}	break;
	case PH_DEACTIVATE_IND:		{pn="DEACTIVATE IND";		dir=1;}	break;
	case PH_DEACTIVATE_REQ:		{pn="DEACTIVATE REQ";		dir=0;}	break;
	case PH_DEACTIVATE_CNF:		{pn="DEACTIVATE CNF";		dir=2;}	break;
	case MPH_ACTIVATE_IND:		{pn="MPH ACTIVATE IND";		dir=1;}	break;
	case MPH_ACTIVATE_REQ:		{pn="MPH ACTIVATE REQ";		dir=0;}	break;
	case MPH_INFORMATION_REQ:	{pn="MPH INFORMATION REQ";	dir=0;}	break;
	case MPH_DEACTIVATE_IND:	{pn="MPH DEACTIVATE IND";	dir=1;}	break;
	case MPH_DEACTIVATE_REQ:	{pn="MPH DEACTIVATE REQ";	dir=0;}	break;
	case MPH_INFORMATION_IND:	{pn="MPH INFORMATION IND";	dir=1;}	break;
	case PH_CONTROL_REQ:		{pn="PH CONTROL REQ";		dir=0;}	break;
	case PH_CONTROL_IND:		{pn="PH CONTROL IND";		dir=1;}	break;
	case PH_CONTROL_CNF:		{pn="PH CONTROL CNF";		dir=2;}	break;
	default:
	  sprintf(pns, "Unknown %04x", hh->prim);
	  pn = pns;
	  break;
      }
      colour="\x1b[;34;1m ";
      if (dir==1) colour="\x1b[;32;1m>";
      if (dir==0) colour="\x1b[;33;1m<";
      printf("%s%s\n", colour, pn);
#endif
      if (result > MISDN_HEADER_LEN)
	dlogframe(buffer+MISDN_HEADER_LEN, result - MISDN_HEADER_LEN, (hh->prim & 1) ^ 1, time);
      fflush(stdout);
    }
  }
  close(log_socket);
}
