/*
 * tracI.c	code to decode ITU Q.931 call control messages
 *
 * Author	Jan den Ouden
 *
 * Changelog
 *
 * Pauline Middelink	general improvements
 *
 * Beat Doebeli		cause texts, display information element
 *
 * Karsten Keil		cause texts, display information element for 1TR6
 *
 *
 * Birger Harzenetter	cause texts, message types, onformation elements, decoding of:
 *	"WIMPy"		type of number, numbering plan, presentation indicator,
 *			screening indicator, date/time, connected number, causes,
 *			progress indicator, redirecting number, redirection reason,
 *			bearer capability, transfer mode, transfer speed, signal,
 *			redirection number, notify, LLC octet 5, 5a, 5b, 5c, 5d
 *			call state, Feature activation, Feature status, Keypad,
 *			Ackermann Euracom PaBXs status messages
 *			some very very basic decoding of facility and extended facility
 *
 *			Added the following additional features:
 *			- Write a call log file
 *			- broadcast caller information to the network
 *			- Phonebook for the two features above
 *			- Additionally now also reads traces from:
 *					- Ascend/Lucent MAX(text file)
 *					- Cisco AS(text file)
 *					- Asterisk(text file)
 *					- mISDNdebugtool and misdn_log(text file)
 *					- Auerswald PaBXs (TCP connection, option -a)
 *					- direct reading from mISDN socket (option -m)
 *					- onsoft watchS0 isdn-analyuzer
 *
 * $Log: q931.c,v $
 * Revision 1.5.2.3  1998/11/03 00:07:28  keil
 * certification related changes
 * fixed logging for smaller stack use
 *
 * Revision 1.5.2.2  1998/10/25 18:16:32  fritz
 * Replaced some read-only variables by defines.
 *
 * Revision 1.5.2.1  1997/10/17 22:14:20  keil
 * update to last hisax version
 *
 * Revision 1.6  1997/07/27 21:09:44  keil
 * move functions to isdnl3.c
 *
 * Revision 1.5  1997/04/06 22:56:43  keil
 * Some cosmetic changes
 *
 * Revision 1.4  1997/02/09 00:29:11  keil
 * new interface handling, one interface per card
 *
 * Revision 1.3  1997/01/21 22:24:59  keil
 * cleanups
 *
 * Revision 1.2  1996/10/27 22:12:45  keil
 * reporting unknown level 3 protocol ids
 *
 * Revision 1.1  1996/10/13 20:04:56  keil
 * Initial revision
 *
 *
 */

#define VERSION "tracI v0.24:\n"

#define INBSIZE 256

#include <ctype.h>
#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <strings.h>
#include <unistd.h>

#ifdef P_AUERSWALD
#include "auerswald.h"
#endif

#ifdef MISDN
#include "misdn.h"
#endif

#include "global.h"


void chkempty(char * name, char * arg) {
  if (*arg!=0) fprintf(stderr, "'%s' does not take any parameter - ignored\n", name);
}


void main_file(char *readfnp) {
  char buf[3*INBSIZE], *bend, time[20]="--?--";
  u_char frame[INBSIZE];
  char *i, *i2, *n;
  int len=-1, dir=0, Type, NextType, Multi=0, d=0;
  FILE *infile;

  fprintf(stderr, "reading data from %s\n", readfnp);

  if (strcmp(readfnp,"-")==0) {
    infile=stdin;
  } else {
    if (!(infile=fopen(readfnp, "r"))) {
      fprintf(stderr,"Unable to read input file \"%s\" (%s)\n", readfnp, strerror(errno));
      exit(1);
    }
  }

  NextType=0;
  bzero(buf, 3*INBSIZE);
  while((fgets(buf, 3*INBSIZE, infile)) != NULL) {
    bend=buf+strlen(buf);
    Type=0;
#ifdef DEBUG
  fprintf(stderr, "\nread(%d)='%s'\n",bend-buf,buf);
#endif

    Type=NextType;
    NextType=0;
    /* I4L WIMPy */
    if (strncmp(buf+9,"!DATA"	,5)==0)			Type=1;
    /* MAX */
    if (strncmp(buf,"BRI-RCV-"	,8)==0)			{Type=2;d=1;}
    if (strncmp(buf,"BRI-XMIT-"	,9)==0)			{Type=2;d=0;}
    if (strncmp(buf,"PRI-RCV-"	,8)==0)			{Type=2;d=1;}
    if (strncmp(buf,"PRI-XMIT-"	,9)==0)			{Type=2;d=0;}
    if (strncmp(buf,"  [00"	,5)==0 && len>=0)	Type=3;
    /* Asterisk zap pri */
    if (strncmp(buf,"< [ "	,4)==0)			{Type=4;d=1;}
    if (strncmp(buf,"> [ "	,4)==0)			{Type=4;d=0;}
    /* Asterisk dahdi pri console */
    if ((n=strstr(buf,"chan_dahdi.c: "))==0) n=strstr(buf,"PRI Span: ");
    if (n) {
      if ((i=strstr(n," < [ "))) {
	if (*buf=='[') strncpy(time,buf+8,8);
	time[8]=0;
	memmove(buf,i+1,strlen(i));
	Type=4;d=1;
      }
      if ((i=strstr(n," > [ "))) {
	if (*buf=='[') strncpy(time,buf+8,8);
	time[8]=0;
	memmove(buf,i+1,strlen(i));
	Type=4;d=0;
      }
    }
    /* mISDNdebugtool */
    if (strncmp(buf,"Received packet from ",21)==0 && (i=strstr(buf," type:D_"))) {
      if (i[8]=='R') d=1; else d=0;
      Type=-1;
      NextType=5;
    }
    /* misdn_log */
    if ((i=strstr(buf," bytes prim = "))) {
      dir=(strtol(i+14, (char **) NULL, 16) & 1) ^ 1;
      Type=6;
      strncpy(time,buf+11,12);
      time[12]=0;
      i--;
      while(*i--==' ');
      while(*i-->' ');
      Multi=atoi(i)-8;
      i=strstr(i," id=");
      i++;
      while(*i++>' ');
      memmove(buf, i, strlen(buf)-(i-buf)+1);
      bend=buf+strlen(buf);
    }
    /* misdn_log */
    if ((i=strstr(buf,"DATA CNF "))) Type=-1;
    i=strstr(buf," DATA REQ ");
    n=strstr(buf," DATA IND ");
    if (n||i) {
      if (n) dir=1; else dir=0;
      strncpy(time,buf+11,12);
      time[12]=0;
      i=strstr(buf," bytes ")-1;
      n=i+7;
      while(*i--==' ');
      while(*i-->' ');
      Multi=atoi(i);
      while(*n++>' ');
      memmove(buf, n, strlen(buf)-(n-buf)+1);
      bend=buf+strlen(buf);
      Type=6;
    }
    if (strncmp(buf+21," NT->TE L",9)==0) {
      if ((i=strstr(buf, " hex:"))) {
        Type=8;dir=0;
        strncpy(time,buf+9,12);
      }
    }
    if (strncmp(buf+21," TE->NT L",9)==0) {
      if ((i=strstr(buf, " hex:"))) {
        Type=8;dir=1;
        strncpy(time,buf+9,12);
      }
    }
#ifdef DEBUG
  fprintf(stderr, "\n>>Type=%d	Time='%s'\n\n",Type,time);
#endif

    switch (Type) {
      case -1:
	/* ignore */
	break;
      case 1:
	dir=buf[14]=='<';
	buf[8]=0;
	len=0;
	i=buf+16;
	while(*i>31) {
	  frame[len++]=strtol(i,&n,16);
	  i=n+1;
	}
	dlogframe(frame,len,dir,buf);
	len=-1;
	break;
      case 2:
	Type=strtol(buf+11,NULL,10);
	if (len>=0 && Type<=Multi) {
	  dlogframe(frame,len,dir,time);
	  len=-1;
	}
	Multi=Type;
	dir=d;
	n=buf+10;
	while(*n!=':') n++;
	while(*n!=':') n++;
	while(*n!='.') n++;
	strncpy(time, n-5, 8);
	time[8]=' ';
	if (Multi<2) len=0;
	break;
      case 3:
	i=buf+10;
	while(*i>31) {
	  frame[len++]=strtol(i,&n,16);
	  i=n+1;
	}
	break;
      case 4:
	dir=d;
	len=0;
	i=buf+4;
	while(*i>31 && *i!=93) {
	  frame[len++]=strtol(i,&n,16);
	  i+=3;
	}
	dlogframe(frame,len,dir,time);
	len=-1;
	break;
      case 5:
	dir=d;
	len=0;
	i=buf;
	while(*i>' ') i++;
	i++;
	while(*i>31) {
	  frame[len++]=strtol(i,&n,16);
	  i+=3;
	}
	dlogframe(frame,len,dir,buf);
	len=-1;
	break;
      case 6:	/* misdn2 */
	len=0;
	i=buf;
	while(*i>31 && i<bend) {
	  frame[len++]=strtol(i,&n,16);
	  i=n+1;
	  Multi--;
	}
	if (Multi) NextType=7;
	else if (len>0) dlogframe(frame,len,dir,time);
	break;
      case 7:	/* misdn2 multiline */
	i=buf;
	while (*i++==' ');
	i--;
	while(*i>31 && i<bend) {
	  frame[len++]=strtol(i,&n,16);
	  i=n+1;
	  Multi--;
	}
	if (Multi) NextType=7;
	else dlogframe(frame,len,dir,time);
	break;
      case 8:	/* OnSoft Watch S0 */
	if (buf[30]=='2') {
	  len=0;
	  i+=5;
	  while(*i>31 && i<bend) {
	    frame[len++]=strtol(i,&n,16);
	    i=n+1;
	  }
	  if (frame[2]&1) {
	    dlogframe(frame,len,dir,time);
	    Multi=0;
	  } else {
	    Multi=len;
	  }
	  len=-1;
	} else if (buf[30]=='3') {
	  i+=5;
	  if (Multi>0) {
	    len=Multi;
	  } else {
	    frame[0]=0;
	    frame[1]=0;
	    frame[2]=0;
	    frame[3]=0;
	    i2=strstr(buf," tei:");
	    if (i2) {
	      d=strtol(i2+5,&n,16);
	      frame[1]=d<<1|1;
	    }
	    len=4;
	  }
	  while(*i>31 && i<bend) {
	    frame[len++]=strtol(i,&n,16);
	    i=n+1;
	  }
	  dlogframe(frame,len,dir,time);
	  len=-1;
	} else {
	  printf("\n?%s\n",i);
	}
	break;
      default:
	if (len>=0) {
	  dlogframe(frame,len,dir,time);
//	  printf("%c[;m",27);
	}
	if (_Unk) {
	  if (len<0) printf("%c[;m",27);
	  printf("%s",buf);
	}
	len=-1;
    }
    fflush(stdout);
    bzero(buf, 3*INBSIZE);
  }
  if (len>=0) {
    dlogframe(frame,len,dir,time);
//    printf("%c[;m",27);
  }
  fclose(infile);
/*  if (n<0) {
    perror("Error reading file");
    exit(2);
  }*/
}


void readconf(char *file) {
  FILE *conf;
  char line[1025], *arg;

  if ((conf=fopen(file,"r"))) {
    fprintf(stderr, "reading configuration from '%s'\n", file);
    while (fgets(line, 1024, conf)!=0) {
      arg=line+strlen(line);
      arg-=1;
      if (*arg==10) *arg=0;
      arg=line;
      while((*arg=tolower(*arg))>32) arg++;
      if ((*(arg))>0) {
	*(arg)=0;
	while(*arg++<33);
	arg--;
      }
#ifdef DEBUG
  fprintf(stderr, "Readconf: key '%s', arg '%s'\n", line, arg);
#endif
      if (*line==0 || *line=='#') ;
      else if (strcmp(line,"nosframes")==0)	{_L2=0;chkempty("NoLayer2", arg);}
      else if (strcmp(line,"nounknown")==0)	{_Unk=0;chkempty("NoLayer2", arg);}
      else if (strcmp(line,"readfile"	)==0)	strncpy(readfn,arg,1023);
      else if (strcmp(line,"execonconn"	)==0)	strncpy(execonconn,arg,1023);
      else if (strcmp(line,"execondisc"	)==0)	strncpy(execondisc,arg,1023);
#ifdef U_CTI
      else if (strcmp(line,"intprefix")==0)	strncpy(P_INT,arg,7);
      else if (strcmp(line,"natprefix")==0)	strncpy(P_NAT,arg,7);
      else if (strcmp(line,"locprefix")==0)	strncpy(P_SUB,arg,15);
      else if (strcmp(line,"logfile"	)==0)	strncpy(logfile,arg,1023);
      else if (strcmp(line,"phonebook")==0)	strncpy(phonebook,arg,1023);
      else if (strcmp(line,"sendto"	)==0)	readbc(arg);
      else if (strcmp(line,"autoadd"	)==0)	{autoadd=1;chkempty("AutoAdd", arg);}
      else if (strcmp(line,"nodata"	)==0)	{NoData=1;chkempty("NoData", arg);}
#endif
      else fprintf(stderr, "Unknown configuration keyword '%s'\n",line);
    }
    fclose(conf);
    P_INT[7]=0;
    P_NAT[7]=0;
    P_SUB[15]=0;
  } else	fprintf(stderr, "Unable to open configuration file (%s)\n", strerror(errno));
}

int main(int argc, char **argv) {
  char *conffn=NULL, *readfnp=NULL;
#ifdef MISDN
  char *mport=NULL;
#endif
  char *auhost=NULL;
  int opt, err=0;

  printf(VERSION);
  bzero(CallRef,sizeof(CallRef));

  opterr=0;
#ifdef MISDN
  while ((opt = getopt(argc, argv, "c:i:m:a:")) !=-1) {
#else
  while ((opt = getopt(argc, argv, "c:i:a:")) !=-1) {
#endif
    switch (opt) {
      case 'c':
	if (conffn) err=1;
	conffn=optarg;
	break;
      case 'i':
	if (readfnp) err=1;
	readfnp=optarg;
	break;
#ifdef MISDN
      case 'm':
	if (mport) err=1;
	mport=optarg;
	break;
#endif
#ifdef P_AUERSWALD
      case 'a':
	if (auhost) err=1;
	auhost=optarg;
	break;
#endif
      case '?':
	err=1;
      default:
	err=1;
    }
  }
  if (argc>optind && !conffn) conffn=argv[optind++];
  if (argc>optind) err=1;
#ifdef MISDN
  if (readfnp && mport) err=1;
#ifdef P_AUERSWALD
  if (auhost && mport) err=1;
#endif
#endif
  if (readfnp && auhost) err=1;
  if (err) {
    fprintf(stderr, "invalid options\n");
    fprintf(stderr, "usage:\n");
    fprintf(stderr, "tracI [[-c] configfile] [-i inputfile"
#ifdef MISDN
    "|-m misnport"
#endif
#ifdef P_AUERSWALD
    "|-a host"
#endif
    "]\n");
    exit(2);
  }
  if (!conffn) conffn=".tracIrc";

  readconf(conffn);
  if (!readfnp) readfnp=readfn;

#ifdef MISDN
  if (mport) main_mport(atol(mport));
#endif
#ifdef P_AUERSWALD
  if (auhost) main_auerswald(auhost);
#endif
  main_file(readfnp);
  return(0);
}
